# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from kqcircuits.elements.element import Element
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.defaults import default_bump_type, default_bump_parameters
from kqcircuits.elements.flip_chip_connectors import connector_type_choices


class FlipChipConnector(Element):
    """Connector between matching faces of two chips.

    The connector makes a galvanic contact between the flipped over top chip and the bottom chip.
    Origin is at the geometric center.
    """

    default_type = default_bump_type
    ubm_diameter = Param(
        pdt.TypeDouble, "Under-bump metalization diameter", default_bump_parameters["under_bump_diameter"], unit="μm"
    )
    bump_diameter = Param(pdt.TypeDouble, "Bump diameter", default_bump_parameters["bump_diameter"], unit="μm")
    bump_type = Param(pdt.TypeString, "Bump type", default_bump_type, choices=connector_type_choices)

    @classmethod
    def create(cls, layout, library=None, bump_type=None, **parameters):
        """Create a bump cell in layout."""
        return cls.create_subtype(layout, library, bump_type, **parameters)[0]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from kqcircuits.elements.flip_chip_connectors.flip_chip_connector import FlipChipConnector
from kqcircuits.util.geometry_helper import circle_polygon


class FlipChipConnectorDc(FlipChipConnector):
    """PCell declaration for an inter-chip dc connector.

    .. MARKERS_FOR_PNG 0,0 0,15
    """

    def build(self):

        ubm_shape = circle_polygon(self.ubm_diameter / 2, self.n)
        self.cell.shapes(self.get_layer("underbump_metallization", 0)).insert(ubm_shape)
        self.cell.shapes(self.get_layer("underbump_metallization", 1)).insert(ubm_shape)

        avoidance_shape = circle_polygon(self.ubm_diameter / 2 + self.margin, self.n)
        self.cell.shapes(self.get_layer("ground_grid_avoidance", 0)).insert(avoidance_shape)
        self.cell.shapes(self.get_layer("ground_grid_avoidance", 1)).insert(avoidance_shape)

        bump_shape = circle_polygon(self.bump_diameter / 2, self.n)
        self.cell.shapes(self.get_layer("indium_bump", 0)).insert(bump_shape)  # bottom In bump
        self.cell.shapes(self.get_layer("indium_bump", 1)).insert(bump_shape)  # top In bump



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from kqcircuits.util.geometry_helper import bspline_points
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.qubits.double_pads import DoublePads
from kqcircuits.pya_resolver import pya


class DoublePadsSplines(DoublePads):
    """A two-island qubit, consisting of two spline-constructed islands shunted by a junction,
    with one capacitive coupler.

    Contains a coupler on the north edge and two separate qubit islands in the center
    joined by a junction or SQUID loaded from another library.
    Refpoint for a readout line at the opening to the coupler and a modifiable refpoint for
    a driveline outside of the rectangle.
    """

    island_spline = Param(
        pdt.TypeList,
        "Control points of a B-Spline that defines left half of the island",
        [-250, 20, -250, 80, -100, 100, 0, 100],
    )
    island_spline_samples = Param(
        pdt.TypeInt,
        "Number of samples taken from each island spline",
        100,
        docstring=(
            "Number of samples taken from each island spline. "
            "There is a spline for each consequent series of four control points"
        ),
    )

    def _build_island1(self, squid_height, taper_height):
        island1_bottom = self.squid_offset + squid_height / 2
        curve_start = pya.DPoint(0, island1_bottom + taper_height)
        island1_control_points = [curve_start + pya.DPoint(-self.island1_taper_width / 2, 0)] + [
            curve_start + pya.DPoint(float(self.island_spline[idx]), float(self.island_spline[idx + 1]))
            for idx in range(0, len(self.island_spline), 2)
        ]
        island1_control_points += [pya.DPoint(-p.x, p.y) for p in reversed(island1_control_points[:-1])]
        for i, p in enumerate(island1_control_points):
            self.refpoints[f"island1_control_point_{i}"] = p
        island1_polygon = pya.DPolygon(
            bspline_points(island1_control_points, self.island_spline_samples, startpoint=True, endpoint=True)
        )
        island1_region = pya.Region(island1_polygon.to_itype(self.layout.dbu))
        island1_taper = pya.Region(
            pya.DPolygon(
                [
                    # pertrude taper into island in case of precision errors
                    pya.DPoint(self.island1_taper_width / 2, island1_bottom + taper_height + 1),
                    pya.DPoint(self.island1_taper_width / 2, island1_bottom + taper_height),
                    pya.DPoint(self.island1_taper_junction_width / 2, island1_bottom),
                    pya.DPoint(-self.island1_taper_junction_width / 2, island1_bottom),
                    pya.DPoint(-self.island1_taper_width / 2, island1_bottom + taper_height),
                    # pertrude taper into island in case of precision errors
                    pya.DPoint(-self.island1_taper_width / 2, island1_bottom + taper_height + 1),
                ]
            ).to_itype(self.layout.dbu)
        )
        return island1_region + island1_taper

    def _build_island2(self, squid_height, taper_height):
        island2_top = self.squid_offset - squid_height / 2
        curve_start = pya.DPoint(0, island2_top - taper_height)
        island2_control_points = [curve_start + pya.DPoint(-self.island2_taper_width / 2, 0)] + [
            curve_start + pya.DPoint(float(self.island_spline[idx]), -float(self.island_spline[idx + 1]))
            for idx in range(0, len(self.island_spline), 2)
        ]
        island2_control_points += [pya.DPoint(-p.x, p.y) for p in reversed(island2_control_points[:-1])]
        for i, p in enumerate(island2_control_points):
            self.refpoints[f"island2_control_point_{i}"] = p
        island2_polygon = pya.DPolygon(
            bspline_points(island2_control_points, self.island_spline_samples, startpoint=True, endpoint=True)
        )
        island2_region = pya.Region(island2_polygon.to_itype(self.layout.dbu))
        island2_taper = pya.Region(
            pya.DPolygon(
                [
                    # pertrude taper into island in case of precision errors
                    pya.DPoint(self.island2_taper_width / 2, island2_top - taper_height - 1),
                    pya.DPoint(self.island2_taper_width / 2, island2_top - taper_height),
                    pya.DPoint(self.island2_taper_junction_width / 2, island2_top),
                    pya.DPoint(-self.island2_taper_junction_width / 2, island2_top),
                    pya.DPoint(-self.island2_taper_width / 2, island2_top - taper_height),
                    # pertrude taper into island in case of precision errors
                    pya.DPoint(-self.island2_taper_width / 2, island2_top - taper_height - 1),
                ]
            ).to_itype(self.layout.dbu)
        )
        return island2_region + island2_taper



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from importlib import import_module
from io import IOBase
from pathlib import Path
from sys import platform
import os
import site
import setuptools


def install_kqc_gui_dependencies():
    """Check KQCircuits' dependencies against klayout-requirements.txt file and install/upgrade if missing.

    This is *only* for KLayout GUI. Stand-alone mode needs manual pip install or pip-sync, preferably in a venv.
    This function should run only once at KLayout startup.
    """
    # pylint: disable=import-outside-toplevel

    from kqcircuits.pya_resolver import pya

    # Skip installation in stand-alone python package mode
    if not hasattr(pya, "MessageBox"):
        return

    detected_os = None
    if os.name == "nt":  # Windows
        detected_os = "win"
    elif os.name == "posix":
        if platform == "darwin":  # macOS
            detected_os = "mac"
        else:
            detected_os = "linux"
    else:
        raise SystemError("Unsupported operating system.")

    target_dir = os.path.split(setuptools.__path__[0])[0]
    test_file = IOBase()
    try:
        test_file_path = os.path.join(target_dir, ".test.file")
        # Following throws PermissionError if target_dir needs sudo
        test_file = open(test_file_path, "x")  # pylint: disable=R1732
        test_file.close()
        if os.path.exists(test_file_path):
            os.remove(test_file_path)
    except PermissionError:
        target_dir = site.USER_SITE
    finally:
        test_file.close()

    mismatch = {}
    # Check path expected after Developer guide installation
    requirements_dir = Path(f"{os.path.dirname(__file__)}/../../kqcircuits_requirements")
    if not requirements_dir.exists():
        # Check path expected after SALT installation
        requirements_dir = Path(f"{os.path.dirname(__file__)}/../../requirements")
        if not requirements_dir.exists():
            raise FileNotFoundError(
                "Can't find gui-requirements.txt file. "
                + "If you used a developer GUI setup, try running 'python3 setup_within_klayout.py'"
            )
    requirements_file = f"{requirements_dir}/{detected_os}/gui-requirements.txt"
    with open(requirements_file) as f:
        for line in f:
            line = line.split("#")[0]
            tokens = line.split("==")
            if len(tokens) < 2:
                continue
            package = tokens[0].strip()
            version = tokens[1].split("\\")[0].strip()
            try:
                mod = import_module(package)
                if not hasattr(mod, "__version__"):
                    raise ModuleNotFoundError()
                if mod.__version__ != version:
                    mismatch[package] = (version, mod.__version__)
            except ModuleNotFoundError:
                mismatch[package] = (version, None)

    if not mismatch:
        return

    mismatch_msg = "\n".join(
        [
            (
                f"  '{package}' ({expected_version}) - not installed"
                if not installed_version
                else f"  '{package}' expected version {expected_version}, got {installed_version}"
            )
            for package, (expected_version, installed_version) in mismatch.items()
        ]
    )

    # Install missing modules inside KLayout.
    from pip import __main__

    if hasattr(__main__, "_main"):
        main = __main__._main
    else:
        from pip._internal.cli.main import main

    ask = pya.MessageBox.warning(
        "Dependencies out of date",
        "Some dependencies for KQCircuits GUI were found to be out of date, according to\n"
        + f"{requirements_file}\n\n"
        + "The affected dependencies are:\n\n"
        + mismatch_msg
        + "\n\nInstall up to date dependencies?",
        pya.MessageBox.Yes + pya.MessageBox.No,
    )
    if ask == pya.MessageBox.Yes:
        main(["install", "-r", requirements_file, "--upgrade", f"--target={target_dir}"])
        error_msg = ""
        for package, (expected_version, _) in mismatch.items():
            try:
                mod = import_module(package)
                if not hasattr(mod, "__version__"):
                    raise ModuleNotFoundError()
                if mod.__version__ != expected_version:
                    error_msg += f"{package} still has version {mod.__version__} instead of {expected_version}\n"
            except ModuleNotFoundError:
                error_msg += f"{package} still not installed\n"
        if error_msg:
            pya.MessageBox.warning(
                "Dependency update not in effect",
                f"{error_msg}\nDependency update has not come into effect, KLayout restart is needed.\n\n"
                + "If a prompt related to dependencies appears again, "
                + "this means something went wrong with the update.\n",
                0,
            )



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import logging

from kqcircuits.chips.chip import Chip
from kqcircuits.elements.airbridge_connection import AirbridgeConnection
from kqcircuits.elements.meander import Meander
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter
from kqcircuits.pya_resolver import pya
from kqcircuits.qubits.circular_transmon_single_island import CircularTransmonSingleIsland
from kqcircuits.qubits.double_pads import DoublePads
from kqcircuits.util.coupler_lib import cap_params
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(Chip, name_chip="EM1")
@add_parameters_from(
    DoublePads,
    coupler_extent=[150, 20],
    island1_extent=[1000, 200],
    island2_extent=[1000, 200],
    island_island_gap=200,
    ground_gap=[1400, 900],
    drive_position=[-1100, 400],
)
class MunchQubits(Chip):
    """Demonstration chip with two circular single island qubits, one floating island qubit, three readout resonators,
    one probe line, three drivelines and one resonant coupler.

    """

    # Readout parameters
    readout_res_lengths = Param(pdt.TypeList, "Readout resonator lengths", [11500, 12700, 8000], unit="[μm]")
    kappa_finger_control = Param(
        pdt.TypeList, "Finger control for the input capacitor", [3.32, 4.21, 1.46], unit="[μm]"
    )
    # Coupling parameters
    coupler_length = Param(pdt.TypeDouble, "Resonant coupler length", 9800, unit="µm")
    # Circular qubit 1 parameters
    couplers_a_qb1 = Param(pdt.TypeList, "Width of the coupler waveguide's center conductors", [10, 3], unit="[μm]")
    couplers_b_qb1 = Param(pdt.TypeList, "Width of the coupler waveguide's gaps", [6, 32], unit="[μm]")
    couplers_angle_qb1 = Param(
        pdt.TypeList,
        "Positioning angles of the couplers, where 0deg corresponds to positive x-axis",
        [225, 315],
        unit="[degrees]",
    )
    couplers_width_qb1 = Param(pdt.TypeList, "Radial widths of the arc couplers", [30, 50], unit="[μm]")
    couplers_arc_amplitude_qb1 = Param(pdt.TypeList, "Couplers angular extension", [25, 65], unit="[degrees]")
    # Circular qubit 2 parameters
    couplers_a_qb2 = Param(pdt.TypeList, "Width of the coupler waveguide's center conductors", [10, 3], unit="[μm]")
    couplers_b_qb2 = Param(pdt.TypeList, "Width of the coupler waveguide's gaps", [6, 32], unit="[μm]")
    couplers_angle_qb2 = Param(
        pdt.TypeList,
        "Positioning angles of the couplers, where 0deg corresponds to positive x-axis",
        [315, 225],
        unit="[degrees]",
    )
    couplers_width_qb2 = Param(pdt.TypeList, "Radial widths of the arc couplers", [30, 50], unit="[μm]")
    couplers_arc_amplitude_qb2 = Param(pdt.TypeList, "Couplers angular extension", [35, 65], unit="[degrees]")

    drive_line_offsets = Param(
        pdt.TypeList, "Distance between the end of a drive line and the qubit pair", [550.0] * 2, unit="[µm]"
    )

    # Floating double pad qubit 3 parameters are added instead as @add_parameters_from the element since there is only
    # one  qubit, they will be passed automatically to the element when added

    def build(self):
        # Define launchpads positioning and function
        launcher_assignments = {
            1: "DL-QB1",
            2: "DL-QB2",
            3: "PL-1-OUT",
            5: "DL-QB3",
            8: "PL-1-IN",
        }
        # Use an 8 port default launcher
        self.produce_launchers("SMA8", launcher_assignments)
        self.produce_qubits()
        self.produce_coupler()
        self.produce_probeline()
        self.produce_readout_resonators()
        self.produce_drivelines()

    def produce_qubits(self):
        # Position the circular qubits
        transformations = [pya.DCplxTrans(1, 0, False, 3500, 7000), pya.DCplxTrans(1, 0, False, 6500, 7000)]
        drive_angles = [110, 70]

        # Make a function to add a single circular qubit
        def produce_circular_qubit(
            name,
            trans,
            couplers_a,
            couplers_b,
            couplers_angle,
            couplers_width,
            couplers_arc_amplitude,
            drive_angle,
            drive_line_offset,
        ):
            qubit_cell = self.add_element(
                CircularTransmonSingleIsland,
                r_island=300,
                ground_gap=200,
                squid_angle=90,
                drive_angle=drive_angle,
                drive_distance=float(drive_line_offset),
                couplers_r=400,
                couplers_a=list(map(float, couplers_a)),
                couplers_b=list(map(float, couplers_b)),
                couplers_angle=list(map(float, couplers_angle)),
                couplers_width=list(map(float, couplers_width)),
                couplers_arc_amplitude=list(map(float, couplers_arc_amplitude)),
            )
            _, _ = self.insert_cell(qubit_cell, trans, name, rec_levels=None)

        # Insert both circular qubits
        produce_circular_qubit(
            "QB1",
            transformations[0],
            self.couplers_a_qb1,
            self.couplers_b_qb1,
            self.couplers_angle_qb1,
            self.couplers_width_qb1,
            self.couplers_arc_amplitude_qb1,
            drive_angles[0],
            self.drive_line_offsets[0],
        )
        produce_circular_qubit(
            "QB2",
            transformations[1],
            self.couplers_a_qb2,
            self.couplers_b_qb2,
            self.couplers_angle_qb2,
            self.couplers_width_qb2,
            self.couplers_arc_amplitude_qb2,
            drive_angles[1],
            self.drive_line_offsets[1],
        )

        # Add now the floating island qubit
        qubit_cell = self.add_element(DoublePads)
        _, _ = self.insert_cell(qubit_cell, pya.DCplxTrans(1, 180, False, 5000, 4000), "QB3", rec_levels=None)

    def produce_coupler(self):
        # Insert a fixed coupler of a variable meander size in between qubits
        _, _, length = WaveguideComposite.produce_fixed_length_waveguide(
            self,
            lambda x: [
                Node(self.refpoints["QB1_port_coupler_2"]),
                Node(self.refpoints["QB1_port_coupler_2_corner"], n_bridges=1),
                Node(pya.DPoint(4500, 6500), n_bridges=2),
                Node(pya.DPoint(5500, 6500), length_before=x, n_bridges=6),
                Node(self.refpoints["QB2_port_coupler_2_corner"], n_bridges=2),
                Node(self.refpoints["QB2_port_coupler_2"], n_bridges=1),
            ],
            initial_guess=5000,
            length=self.coupler_length,
            a=float(self.couplers_a_qb1[1]),
            b=float(self.couplers_b_qb1[1]),
            term1=0,
            term2=0,
        )

        logging.info(f"Coupler line length: {length:.2f}")

    def produce_probeline(self):
        # Make the probeline pass through the resonators tees
        probeline = self.add_element(
            WaveguideComposite,
            nodes=[
                Node(self.refpoints["PL-1-IN_base"]),
                Node(self.refpoints["PL-1-IN_port_corner"], n_bridges=1),
                Node(pya.DPoint(self.refpoints["QB1_base"].x - 1000, 1500), n_bridges=4),
                Node(
                    pya.DPoint(self.refpoints["QB1_base"].x, 1500),
                    WaveguideCoplanarSplitter,
                    align=("port_a", "port_c"),
                    angles=[180, 90, 0],
                    lengths=[50, 150, 50],
                    inst_name="QB1_tee",
                    use_airbridges=True,
                    n_bridges=1,
                ),
                Node(
                    pya.DPoint(5000, 1500),
                    WaveguideCoplanarSplitter,
                    align=("port_a", "port_c"),
                    angles=[180, 90, 0],
                    lengths=[50, 150, 50],
                    inst_name="QB3_tee",
                    use_airbridges=True,
                    n_bridges=2,
                ),
                Node(
                    pya.DPoint(self.refpoints["QB2_base"].x, 1500),
                    WaveguideCoplanarSplitter,
                    align=("port_a", "port_c"),
                    angles=[180, 90, 0],
                    lengths=[50, 150, 50],
                    inst_name="QB2_tee",
                    use_airbridges=True,
                    n_bridges=2,
                ),
                Node(pya.DPoint(self.refpoints["QB2_base"].x + 1000, 1500), n_bridges=1),
                Node(self.refpoints["PL-1-OUT_port_corner"], n_bridges=4),
                Node(self.refpoints["PL-1-OUT_base"], n_bridges=1),
            ],
            a=self.a,
            b=self.b,
            term1=0,
            term2=0,
        )
        self.insert_cell(probeline, inst_name="pl")

    def produce_drivelines(self):
        # Connect the drivelines to the qubit ports
        # Circular qubits
        for qubit_nr in range(1, 3):
            self.insert_cell(
                WaveguideComposite,
                nodes=[
                    Node(self.refpoints[f"DL-QB{qubit_nr}_base"]),
                    Node(self.refpoints[f"DL-QB{qubit_nr}_port_corner"], n_bridges=1),
                    Node(self.refpoints[f"DL-QB{qubit_nr}_port_corner"] + pya.DPoint(0, -1200), n_bridges=1),
                    Node(self.refpoints[f"QB{qubit_nr}_port_drive_corner"], n_bridges=1),
                    Node(self.refpoints[f"QB{qubit_nr}_port_drive"]),
                ],
                term2=self.b,
            )
            # Double pad qubit
            airbridge_crossing_coordinate = (
                self.refpoints["pl_QB2_tee_base"] - self.refpoints["pl_QB3_tee_base"]
            ) / 2 + self.refpoints["pl_QB3_tee_base"]
            self.insert_cell(
                WaveguideComposite,
                nodes=[
                    Node(self.refpoints["DL-QB3_base"]),
                    Node(self.refpoints["DL-QB3_port_corner"], n_bridges=1),
                    Node(airbridge_crossing_coordinate + pya.DVector(0, -300), n_bridges=1),
                    Node(airbridge_crossing_coordinate, AirbridgeConnection, n_bridges=1),
                    Node(airbridge_crossing_coordinate + pya.DVector(0, 200), n_bridges=1),
                    Node(
                        pya.DPoint(
                            self.refpoints["QB3_port_drive_corner"].x,
                            (airbridge_crossing_coordinate + pya.DVector(0, 300)).y,
                        ),
                    ),
                    Node(self.refpoints["QB3_port_drive_corner"], n_bridges=2),
                    Node(self.refpoints["QB3_port_drive"]),
                ],
                term2=self.b,
            )

    def produce_readout_resonators(self):
        # Break down the resonator in few parts for simplicity
        tee_angles = [90, 90, 90]  # Coupling port direction at the probeline
        for i, t_angle in enumerate(tee_angles):
            capacitor = self.add_element(
                **cap_params(
                    fingers=float(self.kappa_finger_control[i]),
                    coupler_type="smooth",
                    element_key="cls",
                    fixed_length=160,
                )
            )
            _, cplr_ref = self.insert_cell(
                capacitor,
                trans=pya.DCplxTrans(1, t_angle, False, 0, 0),
                align_to=f"pl_QB{i + 1}_tee_port_b",
                align="port_a",
            )
            # Add the lower part of the resonator, align it, measure it
            # Qubits ports are called slightly different for the circular qubits and the double pad
            qubits_couplers_corners = [f"QB{i + 1}_port_coupler_1_corner"] * 2 + [f"QB{i + 1}_port_cplr_corner"]
            qubits_couplers_refp = [f"QB{i + 1}_port_coupler_1"] * 2 + [f"QB{i + 1}_port_cplr"]
            resonator_bottom, _ = self.insert_cell(
                WaveguideComposite,
                nodes=[
                    Node(cplr_ref["port_b"]),
                    Node(cplr_ref["port_b_corner"]),
                    Node(pya.DPoint(self.refpoints[qubits_couplers_corners[i]].x, cplr_ref["port_b_corner"].y + 100)),
                    Node(
                        pya.DPoint(self.refpoints[qubits_couplers_corners[i]].x, cplr_ref["port_b_corner"].y + 200),
                        n_bridges=1,
                    ),
                ],
                inst_name=f"resonator_bottom_{i + 1}",
            )
            length_nonmeander_bottom = resonator_bottom.cell.length()
            # Add the upper part of the resonator, align it, measure it
            resonator_top, _ = self.insert_cell(
                WaveguideComposite,
                nodes=[
                    Node(self.refpoints[qubits_couplers_refp[i]]),
                    Node(self.refpoints[qubits_couplers_corners[i]]),
                    Node(self.refpoints[qubits_couplers_corners[i]] + pya.DPoint(0, -300), n_bridges=1),
                ],
                inst_name=f"resonator_top_{i + 1}",
            )
            length_nonmeander_top = resonator_top.cell.length()
            # Add the missing part in the center in the correct length
            meander, _ = self.insert_cell(
                Meander,
                start=self.refpoints[qubits_couplers_corners[i]] + pya.DPoint(0, -300),
                end=self.refpoints[f"resonator_bottom_{i + 1}_port_b"],
                length=float(self.readout_res_lengths[i]) - length_nonmeander_top - length_nonmeander_bottom,
                n_bridges=18,
            )
            logging.info(
                f"Resonator QB{i + 1} length: "
                f"{length_nonmeander_bottom + length_nonmeander_top + meander.cell.length()}"
            )



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


import logging

from kqcircuits.elements.element import Element
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector import FlipChipConnector
from kqcircuits.elements.launcher import Launcher
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(FingerCapacitorSquare, "a2", "b2")
@add_parameters_from(FlipChipConnector, bump_type="Flip Chip Connector Dc")
class FlipChipConnectorRf(Element):
    """PCell declaration for an inter-chip rf connector.

    Flip chip connector with two coplanar waveguide connections and different ground bump configurations.
    The input port is on the first face and on the left. The output port is on the second face and rotated as chosen.

    .. MARKERS_FOR_PNG 0,0 15,0 0,-40 -28,0
    """

    inter_bump_distance = Param(pdt.TypeDouble, "Distance between In bumps", 100, unit="μm")
    output_rotation = Param(pdt.TypeDouble, "Rotation of output port w.r.t. input port", 180, unit="degrees")
    connector_a = Param(pdt.TypeDouble, "Conductor width at the connector area", 40, unit="μm")
    connector_b = Param(pdt.TypeDouble, "Gap width at the connector area", 40, unit="μm")
    round_connector = Param(pdt.TypeBoolean, "Use round connector shape", False)
    n_center_bumps = Param(pdt.TypeInt, "Number of center bumps in series", 1)

    def build(self):

        # Flip-chip bump
        bump = self._get_bump()
        for i in range(self.n_center_bumps):
            self.insert_cell(bump, pya.DTrans((i - (self.n_center_bumps - 1) / 2) * self.inter_bump_distance, 0))
        bump_ref = self.get_refpoints(bump)
        logging.debug("bump_ref: %s", bump_ref)

        a2 = self.a2 if self.a2 >= 0 else self.a
        b2 = self.b2 if self.b2 >= 0 else self.b

        if self.round_connector:
            # Rounded geometry
            def rounded_plate(center_x, width, length):
                reg = pya.Region(
                    pya.DBox(center_x - length / 2, -width / 2, center_x + length / 2, width / 2).to_itype(
                        self.layout.dbu
                    )
                )
                return reg.rounded_corners(0, min(width, length) / (2 * self.layout.dbu), self.n)

            def trace(start_x, width, length):
                return pya.Region(pya.DBox(start_x, -width / 2, start_x + length, width / 2).to_itype(self.layout.dbu))

            w = self.connector_a + 2 * self.connector_b
            bumps_length = (self.n_center_bumps - 1) * self.inter_bump_distance
            length = w + bumps_length

            def produce_shape(face, a, b, rotation, trace_rotation):
                # Base metal gap outline
                trace_dtrans = pya.DCplxTrans(1, trace_rotation, False, -bumps_length / 2, 0)
                trace_itrans = trace_dtrans.to_itrans(self.layout.dbu)
                region = rounded_plate(0, w, length) + trace(0, a + 2 * b, -w / 2).transformed(trace_itrans)

                # Ground grid avoidance
                avoid_region = region.sized(self.margin / self.layout.dbu, self.margin / self.layout.dbu, 2)

                # Subtract circles under bumps
                for i in range(self.n_center_bumps):
                    region -= rounded_plate(
                        (i - (self.n_center_bumps - 1) / 2) * self.inter_bump_distance,
                        self.connector_a,
                        self.connector_a,
                    )

                # Subtract input trace with possible rotation
                region -= trace(0, a, -w / 2).transformed(trace_itrans)

                # Subtract trace connecting the series bumps
                region -= trace(-bumps_length / 2, a, bumps_length)

                dtrans = pya.DCplxTrans(1, rotation, False, 0, 0)
                itrans = dtrans.to_itrans(self.layout.dbu)
                self.cell.shapes(self.get_layer("base_metal_gap_wo_grid", face)).insert(region.transformed(itrans))
                self.add_protection(avoid_region.transformed(itrans), face)
                self.add_port(
                    "{}_port".format(self.face_ids[face]),
                    dtrans * pya.DPoint(-bumps_length / 2, 0) + trace_dtrans * dtrans * pya.DVector(-w / 2, 0),
                    trace_dtrans * dtrans * pya.DVector(-1, 0),
                    face,
                )

            produce_shape(0, self.a, self.b, 0, 0)
            produce_shape(1, a2, b2, 180, self.output_rotation - 180)

        else:
            # Taper geometry
            s = self.connector_a + (self.n_center_bumps - 1) * self.inter_bump_distance
            trans = pya.DCplxTrans(1, 0, False, bump_ref["base"] + pya.DPoint(-self.connector_a - s / 2, 0))
            tt = pya.DCplxTrans(1, self.output_rotation, False, 0, 0)
            self.insert_cell(
                Launcher,
                trans,
                self.face_ids[0],
                s=s,
                l=self.connector_a,
                a_launcher=self.connector_a,
                b_launcher=self.connector_b,
                launcher_frame_gap=self.connector_b,
            )

            # If there are many center bumps and output direction is not colinear with the input direction than a simple
            # Launcher won't do. For arbitrary angles we use two Launcher objects one to cover the parallel part, like
            # in the colinear case, but without the tapering part and we add a second short Launcher rotated to the
            # right output direction. Additionally, we use the metal addition layer to get rid of the parts where one
            # Launcher's gap overlaps with the others center conductor.
            add_metal = False
            if int(self.output_rotation) not in (0, 180) and self.n_center_bumps > 1:
                ubm = self.connector_a
                ibm = self.inter_bump_distance
                l = (self.n_center_bumps - 1) * ibm / 2
                pts = [
                    pya.DPoint(l - ibm, ubm / 2),
                    pya.DPoint(l, ubm / 2),
                    pya.DPoint(l, -ubm / 2),
                    pya.DPoint(l - ibm, -ubm / 2),
                ]
                self.cell.shapes(self.get_layer("base_metal_addition", self.face_ids[1])).insert(pya.DPolygon(pts))

                lc = ubm * 1.5
                ts = pya.DCplxTrans(1, 180, False, -lc, 0)
                self.insert_cell(
                    Launcher,
                    ts * trans,
                    None,
                    s=s - lc,
                    l=ubm,
                    a_launcher=self.connector_a,
                    b_launcher=self.connector_b,
                    launcher_frame_gap=self.connector_b,
                    face_ids=[self.face_ids[1], self.face_ids[0]],
                    a=ubm,
                    b=ubm,
                )
                s = ubm
                trans = pya.DCplxTrans(1, 0, False, bump_ref["base"] + pya.DPoint(-ubm - s / 2, 0))
                tt = pya.DCplxTrans(1, self.output_rotation, False, (self.n_center_bumps - 1) / 2 * ibm, 0)
                add_metal = True

            self.insert_cell(
                Launcher,
                tt * trans,
                self.face_ids[1],
                s=s,
                l=self.connector_a,
                a_launcher=self.connector_a,
                b_launcher=self.connector_b,
                launcher_frame_gap=self.connector_b,
                face_ids=[self.face_ids[1], self.face_ids[0]],
                a=a2,
                b=b2,
                add_metal=add_metal,
            )

        self._insert_ground_bumps(bump)

    def _insert_ground_bumps(self, bump):
        # Insert ground bumps
        dist_y = 0.5**0.5 * self.inter_bump_distance
        dist_x = dist_y + self.inter_bump_distance * (self.n_center_bumps - 1) / 2
        self.insert_cell(bump, pya.DCplxTrans(1, 0, False, dist_x, dist_y))
        self.insert_cell(bump, pya.DCplxTrans(1, 0, False, -dist_x, dist_y))
        self.insert_cell(bump, pya.DCplxTrans(1, 0, False, dist_x, -dist_y))
        self.insert_cell(bump, pya.DCplxTrans(1, 0, False, -dist_x, -dist_y))

    def _get_bump(self):
        return self.add_element(FlipChipConnector, face_ids=self.face_ids, bump_type=self.bump_type)

    @classmethod
    def get_sim_ports(cls, simulation):
        return Element.face_changer_waveguides(simulation)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import os
import stat

import logging
from typing import Optional, Union, Sequence, Tuple
from pathlib import Path

from kqcircuits.simulations.export.ansys.ansys_solution import AnsysSolution, get_ansys_solution
from kqcircuits.simulations.export.simulation_export import (
    copy_content_into_directory,
    get_post_process_command_lines,
    get_combined_parameters,
    export_simulation_json,
)
from kqcircuits.simulations.export.simulation_validate import validate_simulation
from kqcircuits.util.export_helper import write_commit_reference_file
from kqcircuits.simulations.export.util import export_layers
from kqcircuits.defaults import ANSYS_EXECUTABLE, ANSYS_SCRIPT_PATHS
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.post_process import PostProcess


def export_ansys_json(simulation: Simulation, solution: AnsysSolution, path: Path):
    """
    Export Ansys simulation into json and gds files.

    Arguments:
        simulation: The simulation to be exported.
        solution: The solution to be exported.
        path: Location where to write json and gds files.

    Returns:
         Path to exported json file.
    """
    if simulation is None or not isinstance(simulation, Simulation):
        raise ValueError("Cannot export without simulation")

    # write .gds file
    gds_file = simulation.name + ".gds"
    gds_file_path = str(path.joinpath(gds_file))
    if not Path(gds_file_path).exists():
        export_layers(
            gds_file_path, simulation.layout, [simulation.cell], output_format="GDS2", layers=simulation.get_layers()
        )
    full_name = simulation.name + solution.name
    # collect data for .json file
    json_data = {
        "name": full_name,
        **solution.get_solution_data(),
        **simulation.get_simulation_data(),
        "gds_file": gds_file,
        "parameters": get_combined_parameters(simulation, solution),
    }

    # write .json file
    json_file_path = str(path.joinpath(full_name + ".json"))
    export_simulation_json(json_data, json_file_path)

    return json_file_path


def export_ansys_bat(
    json_filenames,
    path: Path,
    file_prefix="simulation",
    exit_after_run=False,
    execution_script="scripts/import_and_simulate.py",
    post_process=None,
    use_rel_path=True,
):
    """
    Create a batch file for running one or more already exported simulations.

    Arguments:
        json_filenames: List of paths to json files to be included into the batch.
        path: Location where to write the bat file.
        file_prefix: Name of the batch file to be created.
        exit_after_run: Defines if the Ansys Electronics Desktop is automatically closed after running the script.
        execution_script: The script file to be executed.
        post_process: List of PostProcess objects, a single PostProcess object, or None to be executed after simulations
        use_rel_path: Determines if to use relative paths.

    Returns:
         Path to exported bat file.
    """
    run_cmd = "RunScriptAndExit" if exit_after_run else "RunScript"

    bat_filename = str(path.joinpath(file_prefix + ".bat"))
    with open(bat_filename, "w") as file:
        file.write(
            "@echo off\n"
            r'powershell -Command "Get-Process | Where-Object {$_.MainWindowTitle -like \"Run Simulations*\"} '
            '| Select -ExpandProperty Id | Export-Clixml -path blocking_pids.xml"\n'
            "title Run Simulations\n"
            'powershell -Command "$sim_pids = Import-Clixml -Path blocking_pids.xml; if ($sim_pids) '
            r"{ echo \"Waiting for $sim_pids\"; Wait-Process $sim_pids -ErrorAction SilentlyContinue }; "
            'Remove-Item blocking_pids.xml"\n'
        )

        # Commands for each simulation
        for i, json_filename in enumerate(json_filenames):
            printing = "echo Simulation {}/{} - {}\n".format(
                i + 1, len(json_filenames), str(Path(json_filename).relative_to(path))
            )
            file.write(printing)
            command = '"{}" -scriptargs "{}" -{} "{}"\n'.format(
                ANSYS_EXECUTABLE,
                str(Path(json_filename).relative_to(path) if use_rel_path else json_filename),
                run_cmd,
                str(execution_script),
            )
            file.write(command)

        file.write(get_post_process_command_lines(post_process, path, json_filenames))

    # Make the bat file executable in linux
    os.chmod(bat_filename, os.stat(bat_filename).st_mode | stat.S_IEXEC)

    return bat_filename


def export_ansys(
    simulations: Sequence[Union[Simulation, Tuple[Simulation, AnsysSolution]]],
    path: Path,
    script_folder: str = "scripts",
    file_prefix: str = "simulation",
    exit_after_run: bool = False,
    import_script: str = "import_and_simulate.py",
    post_process: Optional[Union[PostProcess, Sequence[PostProcess]]] = None,
    use_rel_path: bool = True,
    skip_errors: bool = False,
    **solution_params,
) -> Path:
    """
    Export Ansys simulations by writing necessary scripts and json, gds, and bat files.

    Arguments:
        simulations: List of Simulation objects or tuples containing Simulation and Solution objects.
        path: Location where to write export files.
        script_folder: Path to the Ansys-scripts folder.
        file_prefix: Name of the batch file to be created.
        exit_after_run: Defines if the Ansys Electronics Desktop is automatically closed after running the script.
        import_script: Name of import script file.
        post_process: List of PostProcess objects, a single PostProcess object, or None to be executed after simulations
        use_rel_path: Determines if to use relative paths.
        skip_errors: Skip simulations that cause errors. Default is False.

            .. warning::

               **Use this carefully**, some of your simulations might not make sense physically and
               you might end up wasting time on bad simulations.
        solution_params: AnsysSolution parameters if simulations is a list of Simulation objects.

    Returns:
        Path to exported bat file.
    """
    write_commit_reference_file(path)
    copy_content_into_directory(ANSYS_SCRIPT_PATHS, path, script_folder)
    json_filenames = []
    common_sol = None if all(isinstance(s, Sequence) for s in simulations) else get_ansys_solution(**solution_params)
    for sim_sol in simulations:
        simulation, solution = sim_sol if isinstance(sim_sol, Sequence) else (sim_sol, common_sol)
        validate_simulation(simulation, solution)
        try:
            json_filenames.append(export_ansys_json(simulation, solution, path))
        except (IndexError, ValueError, Exception) as e:  # pylint: disable=broad-except
            if skip_errors:
                logging.warning(
                    f"Simulation {simulation.name} skipped due to {e.args}. "
                    "Some of your other simulations might not make sense geometrically. "
                    "Disable `skip_errors` to see the full traceback."
                )
            else:
                raise UserWarning(
                    "Generating simulation failed. You can discard the errors using `skip_errors` in `export_ansys`. "
                    "Moreover, `skip_errors` enables visual inspection of failed and successful simulation "
                    "geometry files."
                ) from e

    return export_ansys_bat(
        json_filenames,
        path,
        file_prefix=file_prefix,
        exit_after_run=exit_after_run,
        execution_script=Path(script_folder).joinpath(import_script),
        post_process=post_process,
        use_rel_path=use_rel_path,
    )



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
import importlib.metadata
import json
import logging
import subprocess
import platform
import sys
import os
import argparse
from sys import argv
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from kqcircuits.elements.element import get_refpoints
from kqcircuits.defaults import (
    default_layers,
    TMP_PATH,
    STARTUPINFO,
    default_probe_types,
    default_probe_suffixes,
    recommended_probe_suffix_mapping,
    VERSION_PATHS,
    DRC_PATH,
    EXPORT_PATH_IDENTIFIER,
)
from kqcircuits.klayout_view import KLayoutView, MissingUILibraryException
from kqcircuits.pya_resolver import pya, is_standalone_session, klayout_executable_command


def _probe_point_coordinate(pos, eu=1e-3, sd=4):
    return {"x": round(pos.x * eu, sd), "y": round(pos.y * eu, sd)}


def _probe_point_to_dpoint(pos, eu=1e-3):
    # this doesn't need to be super precise since we currently only use this to compare distances
    return pya.DPoint(pos["x"] / eu, pos["y"] / eu)


# pylint: disable=dangerous-default-value
def generate_probepoints_json(
    cell: pya.Cell,
    face: str = "1t1",
    flip_face: Optional[bool] = None,
    references: List[str] = ["nw"],
    contact: Optional[Union[Tuple[pya.DPoint, pya.DPoint], List[Tuple[pya.DPoint, pya.DPoint]]]] = None,
) -> Dict:
    """For given cell, collects probepoints from cell's refpoints into a json Dict.

    A refpoint is a probepoint if it

        * contains some string from ``default_probe_types`` and
        * has a suffix from ``default_probe_suffixes``

    Json format consists of {'x': x, 'y': y} type 2d points, in millimeter units.
    The returned json object consists of:

        * an 'alignment' point, which tells the position of the reference marker defined in references marker and
        * 'sites' list. Each entry of the list has a 'west' and 'east' point, and also a unique 'id' as string

    Args:
        * cell: cell from which to collect probepoints
        * face: name of the face from where to collect probepoints
        * flip_face: explicitly specifies if the points should be flipped around the y-axis.
            Can be set to None, in which case will infer whether to flip points from the ``face`` arg
        * references: a list of markers to use as alignment references. String values are one of
            "nw", "ne", "sw", "se". If multiple values supplied, the resulting json will have
            "groups" key on top level, with each group containing the marker string as 'id'
            and its own 'alignment' and 'sites' values, grouping each site to its closest marker.
        * contact: a manually defined contact probe, a tuple of two DPoints.
            Can be None so no "contact" site is added, or can be a list if a different "contact"
            site is needed for each reference
    """
    validations = [
        (cell is None, "Cell is null"),
        (not references, "Can't use empty list of references"),
        (isinstance(contact, tuple) and len(contact) < 2, "Singular contact must be tuple of two DPoints"),
        (
            isinstance(contact, list) and (len(contact) != len(references) or any(len(c) < 2 for c in contact)),
            "List of contacts should define a tuple of two DPoints for each reference",
        ),
    ]
    for check, error_text in validations:
        if check:
            error = ValueError(error_text)
            logging.exception(error_text, exc_info=error)
            raise error

    layout = cell.layout()

    refpoints = get_refpoints(layout.layer(default_layers["refpoints"]), cell)

    # check existence of reference markers
    markers = {}
    for reference in references:
        marker_refpoint = f"{face}_marker_{reference.lower()}"
        if marker_refpoint not in refpoints:
            to_legacy_face_name = {face: "", "1t1": "b", "2b1": "t"}
            legacy_marker_refpoint = f"{to_legacy_face_name[face]}_marker_{reference.lower()}"
            if legacy_marker_refpoint in refpoints:
                marker_refpoint = legacy_marker_refpoint
            else:
                logging.warning(
                    (f"The marker or at least its refpoint {marker_refpoint} " f"is missing in the cell {cell.name}!")
                )
                if pya.DPoint(1500, 8500) not in markers.values():
                    logging.warning(f"Setting marker {marker_refpoint} to DPoint(1500, 8500)")
                    markers[reference.upper()] = pya.DPoint(1500, 8500)
                continue
        markers[reference.upper()] = refpoints[marker_refpoint]

    # if not explicitely stated to flip the face, deduce from face string
    if flip_face is None:
        if len(face) < 2:  # legacy face name
            flip_face = face == "t"
        else:  # current face name
            flip_face = face[1] == "b"
    # get boundaries of the chip dimensions
    matching_layer = [
        l
        for l in layout.layer_infos()
        if l.name in [f"{face}_base_metal_gap_wo_grid", f"{face}*base*metal*gap*wo*grid"]
    ]
    if matching_layer:
        bbox_for_face = cell.dbbox_per_layer(layout.layer(matching_layer[0]))
    else:
        logging.warning(f"No geometry found at layer {face}_base_metal_gap_wo_grid!")
        bbox_for_face = pya.DBox(1500, 1500, 8500, 8500) if flip_face else pya.DBox(0, 0, 10000, 10000)
        logging.warning(f"Assuming chip dimensions are at {bbox_for_face}")
    # define transformation function to apply to each DPoint object
    transform = lambda point: pya.DPoint(point - bbox_for_face.p1)
    # flip top-markers back to top side
    if flip_face:
        flip_origin = pya.DPoint(bbox_for_face.p2.x, bbox_for_face.p1.y)
        transform = lambda point: pya.DPoint(flip_origin.x - point.x, point.y - flip_origin.y)
    markers = {k: transform(v) for k, v in markers.items()}

    # initialize dictionaries for each probe point group
    groups = {}
    for marker_name, marker in markers.items():
        groups[marker_name] = {"alignment": _probe_point_coordinate(marker), "sites": []}

    # first collect sites before grouping them
    sites = []
    for probepoint_name, probepoint in refpoints.items():
        probepoint = transform(probepoint)
        name_type = probepoint_name.split("_")[0]
        # if name_type starts with some probe_type, truncate name_type to be the probe_type
        for probe_type in default_probe_types:
            if name_type.lower().startswith(probe_type.lower()):
                name_type = probe_type.lower()
                break
        # extract suffix if probepoint_name uses one from default_probe_suffixes
        suffixes = [s for s in default_probe_suffixes if probepoint_name.endswith(s)]
        if name_type in default_probe_types and suffixes:
            remove_suffix_tokens = max(len(suffixes[0].split("_")) - 2, 1)
            probe_name = "_".join(probepoint_name.split("_")[:-remove_suffix_tokens])
            # find site with id value as this probepoint prefix
            probepoint_sites = [s for s in sites if s["id"] == probe_name]
            probepoint_entry = _probe_point_coordinate(probepoint)
            # create a new site if none found
            if not probepoint_sites:
                sites.append({"id": probe_name, "west": probepoint_entry})
            else:
                site = probepoint_sites[0]
                if "east" not in site:
                    assign_to_west = probepoint_entry["x"] < site["west"]["x"] or (
                        probepoint_entry["x"] == site["west"]["x"]
                        and recommended_probe_suffix_mapping.get(suffixes[0]) == "west"
                    )
                    if assign_to_west:
                        site["east"] = site["west"]
                        direction = "west"
                    else:
                        direction = "east"
                    site[direction] = probepoint_entry
                    expected_direction = recommended_probe_suffix_mapping.get(suffixes[0])
                    if expected_direction is not None and expected_direction != direction:
                        logging.warning(
                            (
                                f"Probepoint {probepoint_name} was mapped to {direction}, "
                                f"but recommended direction for {suffixes[0]} is {expected_direction}"
                            )
                        )
                else:
                    # limited support for more than two point probing
                    for key in site:
                        if key == "id":
                            continue
                        sites.append(
                            {
                                "east": site[key] if site[key]["x"] > probepoint_entry["x"] else probepoint_entry,
                                "id": f"{probe_name}{suffixes[0]}_{key}",
                                "west": probepoint_entry if site[key]["x"] > probepoint_entry["x"] else site[key],
                            }
                        )

    # sanity check that each site has exactly east and west probe
    for site in sites:
        if set(site.keys()) != {"west", "east", "id"}:
            logging.warning(f"Malformed site object detected: {site}")
            if "east" in site and "west" not in site:
                site["west"] = site["east"]
            elif "west" in site and "east" not in site:
                site["east"] = site["west"]
            elif "east" not in site and "west" not in site:
                site["east"] = {"x": 0.0, "y": 0.0}
                site["west"] = {"x": 0.0, "y": 0.0}

    # reason for sorting is to make the exported json more deterministic
    sites.sort(key=lambda site: site["id"])
    for idx, _ in enumerate(sites):
        sites[idx] = dict(sorted(sites[idx].items()))

    # divide probe points into groups by closest marker (multireference only)
    for site in sites:
        midpoint = {
            "x": (site["west"]["x"] + site["east"]["x"]) / 2.0,
            "y": (site["west"]["y"] + site["east"]["y"]) / 2.0,
        }
        midpoint = _probe_point_to_dpoint(midpoint)
        _, closest_marker = sorted(
            [(refpoint.distance(midpoint), marker) for marker, refpoint in markers.items()], key=lambda x: x[0]
        )[
            0
        ]  # sort by distance, get closest tuple, get marker
        groups[closest_marker]["sites"].append(site)

    # add manual "contact" entries
    if contact is not None:
        if isinstance(contact, tuple):
            contact = [contact] * len(references)
        for idx, group in enumerate(groups.values()):
            contact1, contact2 = contact[idx]
            contact1 = _probe_point_coordinate(transform(contact1))
            contact2 = _probe_point_coordinate(transform(contact2))
            west_is_1 = contact1["x"] < contact2["x"]
            group["sites"].append(
                {
                    "east": contact2 if west_is_1 else contact1,
                    "id": "contact",
                    "west": contact1 if west_is_1 else contact2,
                }
            )

    # find probepoint duplicates (within tolerance) and only keep the one with longer id name
    for group_key, group in groups.items():
        for i, site1 in enumerate(group["sites"]):
            if site1 == {}:
                continue
            for j, site2 in enumerate(group["sites"][i + 1 :]):
                if site2 == {}:
                    continue
                too_close = True
                for side in ["west", "east"]:
                    x1 = site1[side]["x"]
                    x2 = site2[side]["x"]
                    y1 = site1[side]["y"]
                    y2 = site2[side]["y"]
                    if ((x1 - x2) ** 2) + ((y1 - y2) ** 2) > (0.001) ** 2:
                        too_close = False
                if too_close:
                    logging.warning(
                        f"Found two sites '{site1['id']}' and '{site2['id']}' with similar coordinates (respectively)"
                    )
                    logging.warning(
                        f"  west {site1['west']['x']},{site1['west']['y']} = {site2['west']['x']},{site2['west']['y']}"
                    )
                    logging.warning(
                        f"  east {site1['east']['x']},{site1['east']['y']} = {site2['east']['x']},{site2['east']['y']}"
                    )
                    logging.warning(
                        (
                            "  will only keep the site "
                            f"'{site1['id'] if len(site1['id']) > len(site2['id']) else site2['id']}'"
                        )
                    )
                    group["sites"][i + j + 1 if len(site1["id"]) > len(site2["id"]) else i].clear()
                if site1 == {}:
                    break
        # pylint: disable=unnecessary-dict-index-lookup
        groups[group_key]["sites"] = [site for site in group["sites"] if site != {}]

    # remove empty groups
    groups = {k: v for k, v in groups.items() if v["sites"]}

    # leave out groups key if only one group
    if len(groups) == 1:
        return list(groups.values())[0]
    return {"groups": [{"id": name, **group} for name, group in groups.items()]}


def generate_probepoints_from_file(
    cell_file: str,
    face: str = "1t1",
    flip_face: Optional[bool] = None,
    references: List[str] = ["nw"],
    contact: Optional[Union[Tuple[pya.DPoint, pya.DPoint], List[Tuple[pya.DPoint, pya.DPoint]]]] = None,
) -> Dict:
    """For an OAS and GDS file containing a chip at its top cell,
    collects probepoints from cell's refpoints into a json Dict.
    A refpoint is a probepoint if it

        * contains some string from ``default_probe_types`` and
        * has a suffix from ``default_probe_suffixes``

    Json format consists of {'x': x, 'y': y} type 2d points, in millimeter units.
    The returned json object consists of:

        * an 'alignment' point, which tells the position of the reference marker defined in references marker and
        * 'sites' list. Each entry of the list has a 'west' and 'east' point, and also a unique 'id' as string

    Args:
        * cell_file: file path to the OAS or GDS file containing a chip at its top cell
        * face: name of the face from where to collect probepoints
        * flip_face: explicitly specifies if the points should be flipped around the y-axis.
            Can be set to None, in which case will infer whether to flip points from the ``face`` arg
        * references: a list of markers to use as alignment references. String values are one of
            "nw", "ne", "sw", "se". If multiple values supplied, the resulting json will have
            "groups" key on top level, with each group containing the marker string as 'id'
            and its own 'alignment' and 'sites' values, grouping each site to its closest marker.
        * contact: a manually defined contact probe, a tuple of two DPoints.
            Can be None so no "contact" site is added, or can be a list if a different "contact"
            site is needed for each reference
    """
    load_opts = pya.LoadLayoutOptions()
    load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
    view = KLayoutView()
    layout = view.layout
    layout.read(cell_file, load_opts)
    cell = layout.top_cells()[-1]
    return generate_probepoints_json(cell, face, flip_face, references, contact)


def create_or_empty_tmp_directory(dir_name):
    """Creates directory into TMP_PATH or removes its content if it exists.
    Returns directory path.
    """

    def remove_content(path):
        """Removes content of the directory path without removing directory itself."""
        for child in path.iterdir():
            if child.is_dir():
                remove_content(child)
                child.rmdir()
            else:
                child.unlink()

    parser = argparse.ArgumentParser()

    parser.add_argument("--simulation-export-path", type=str, default=None)

    dir_path = get_simulation_directory(dir_name)

    if dir_path.exists() and dir_path.is_dir():
        remove_content(dir_path)
    else:
        dir_path.mkdir()

    # This is used to pass the used export paths back to the `kqc sim` console script
    print(f"{EXPORT_PATH_IDENTIFIER}{dir_path}", flush=True)

    return dir_path


def get_simulation_directory(dir_name):
    """
    Returns directory path consistent with `create_or_empty_tmp_directory`.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("--simulation-export-path", type=str, default=None)
    args, _ = parser.parse_known_args()

    if args.simulation_export_path is not None:
        dir_path = Path(args.simulation_export_path)
    else:
        dir_path = TMP_PATH.joinpath(dir_name)

    return dir_path


def get_active_or_new_layout():
    """Tries to return active layout in GUI or returns new layout when running standalone."""
    try:
        klayoutview = KLayoutView(current=True)
        klayoutview.add_default_layers()
        return klayoutview.layout
    except MissingUILibraryException:
        return pya.Layout()


def write_commit_reference_file(path: Path, write_versions_file=True):
    """
    Writes file COMMIT_REFERENCE into given file path. The file includes current git revision number.
    If git repository is not found in given path, no file is written.
    """
    try:
        with open(path.joinpath("COMMIT_REFERENCE"), "w") as file:
            for item in VERSION_PATHS.items():
                output = subprocess.check_output(
                    ["git", "rev-parse", "HEAD"], stderr=subprocess.DEVNULL, cwd=item[1], startupinfo=STARTUPINFO
                )
                file.write("{} revision number: {}".format(item[0], output.decode("ascii")))

    except subprocess.CalledProcessError:
        return

    if write_versions_file:
        write_export_machine_versions_file(path)


def write_export_machine_versions_file(path: Path):
    """
    Writes file EXPORT_MACHINE_VERSIONS into given file path.
    """
    versions = {}
    versions["platform"] = platform.platform()
    versions["python"] = sys.version_info
    versions["klayout"] = get_klayout_version()

    with open(path.joinpath("EXPORT_MACHINE_VERSIONS.json"), "w") as file:
        json.dump(versions, file)


def open_with_klayout_or_default_application(filepath):
    """
    Tries to open file with KLayout. If KLayout is not found, opens file with operating system's default application.
    Implementation supports Windows, macOS, and Linux.
    """
    if argv[-1] == "-q":  # quiet mode, do not run viewer
        return

    exe = klayout_executable_command()
    if not exe:
        logging.warning("KLayout executable not found.")
    else:
        subprocess.call((exe, filepath))


def get_klayout_version():
    if is_standalone_session():
        return f"KLayout {importlib.metadata.version('klayout')}"
    else:
        return pya.Application.instance().version()


def export_drc_report(name, path, drc_script):
    """Run a DRC script on ``path/name.oas`` and export results in ``path/name_drc_report.lyrdb``."""

    drc_runset_path = os.path.join(DRC_PATH, drc_script)
    if not os.path.exists(drc_runset_path):
        raise FileNotFoundError(f"DRC script {drc_script} does not exist")

    input_file = os.path.join(path, f"{name}.oas")
    output_file = os.path.join(path, f"{name}_drc_report.lyrdb")
    logging.info("Exporting DRC report to %s", output_file)

    try:
        subprocess.run(
            [
                klayout_executable_command(),
                "-b",
                "-i",
                "-r",
                drc_runset_path,
                "-rd",
                f"output={output_file}",
                input_file,
            ],
            check=True,
            startupinfo=STARTUPINFO,
        )
    except subprocess.CalledProcessError as e:
        logging.error(e.output)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from kqcircuits.pya_resolver import pya

DPoint = pya.DPoint


class Port:
    """Base data structure for simulation ports.

    Depending on your simulation type, these produce excitations, set potentials, or act as ideal RLC lumped elements.
    """

    def __init__(
        self,
        number: int,
        resistance: float = 50,
        reactance: float = 0,
        inductance: float = 0,
        capacitance: float = 0,
        face: int = 0,
        junction: bool = False,
        renormalization: float = 50,
    ):
        """
        Args:
            number: Port number.
            resistance: Real part of impedance. Given in Ohms (:math:`\\Omega`).
            reactance: Imaginary part of impedance. Given in Ohms (:math:`\\Omega`).
            inductance: Inductance of the element. Given in Henrys (:math:`\\text{H}`).
            capacitance: Capacitance of the element. Given in Farads (:math:`\\text{F}`).
            face: Integer-valued face index for the port.
            junction: Whether this port models a SQUID/Junction. Used in EPR calculations.
            renormalization: Port renormalization in Ohms or None to not re-normalize the port.
        """
        self.number = number
        self.resistance = resistance
        self.reactance = reactance
        self.inductance = inductance
        self.capacitance = capacitance
        self.face = face
        self.junction = junction
        self.renormalization = renormalization
        self.type = type(self).__name__

    def as_dict(self):
        """Returns attributes as a dictionary."""
        return vars(self)


class InternalPort(Port):
    """Data structure for ports inside the simulation area."""

    def __init__(
        self,
        number: int,
        signal_location: DPoint,
        ground_location: DPoint = None,
        resistance: float = 50,
        reactance: float = 0,
        inductance: float = 0,
        capacitance: float = 0,
        face: int = 0,
        junction: bool = False,
        signal_layer: str = "signal",
        etch_width: float = None,
    ):
        """
        Args:
            number: Port number.
            signal_location: Edge location for signal source.
            ground_location: Edge location to connect signal to. Usually ground.
            resistance: Real part of impedance. Given in Ohms (:math:`\\Omega`).
            reactance: Imaginary part of impedance. Given in Ohms (:math:`\\Omega`).
            inductance: Inductance of the element. Given in Henrys (:math:`\\text{H}`).
            capacitance: Capacitance of the element. Given in Farads (:math:`\\text{F}`).
            face: Integer-valued face index for the port.
            junction: Whether this port models a SQUID/Junction. Used in EPR calculations.
            signal_layer: Manual override for simulation signal layer.
                May be used to set ports across the ground layer with ``ground``.
            etch_width: Width of a trace between signal_location and ground_location, on which the metal is etched away.
                Useful when adding a lumped port on a waveguide.
        """
        super().__init__(number, resistance, reactance, inductance, capacitance, face, junction)
        self.signal_location = signal_location
        self.signal_layer = signal_layer
        if ground_location is not None:
            self.ground_location = ground_location
        if etch_width is not None:
            self.etch_width = etch_width

    def get_etch_polygon(self):
        """Returns polygon under which the metal should be etched away"""
        try:
            d = self.signal_location - self.ground_location
            v = (0.5 * self.etch_width / d.length()) * pya.DVector(-d.y, d.x)
            return pya.DPolygon(
                [self.signal_location - v, self.signal_location + v, self.ground_location + v, self.ground_location - v]
            )
        except AttributeError:
            return pya.DPolygon()


class EdgePort(Port):
    """Data structure for ports at the edge of the simulation area."""

    def __init__(
        self,
        number: int,
        signal_location: DPoint,
        resistance: float = 50,
        reactance: float = 0,
        inductance: float = 0,
        capacitance: float = 0,
        deembed_len: float = None,
        face: int = 0,
        junction: bool = False,
        size=None,
    ):
        """
        Args:
            number: Port number.
            signal_location: Edge location for signal source.
            resistance: Real part of impedance. Given in Ohms (:math:`\\Omega`).
            reactance: Imaginary part of impedance. Given in Ohms (:math:`\\Omega`).
            inductance: Inductance of the element. Given in Henrys (:math:`\\text{H}`).
            capacitance: Capacitance of the element. Given in Farads (:math:`\\text{F}`).
            deembed_len: Port de-embedding length. Given in simulation units, usually microns (:math:`\\text{um}`).
            face: Integer-valued face index for the port.
            junction: Whether this port models a SQUID/Junction. Used in EPR calculations.
            size: Width and height of the port to override Simulation.port_size. Optionally, the size can be set as a
                list specifying the extensions from the center of the port to left, right, down and up, respectively.
        """
        super().__init__(number, resistance, reactance, inductance, capacitance, face, junction)
        self.signal_location = signal_location
        self.deembed_len = deembed_len
        self.size = size



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
from dataclasses import dataclass, field
from typing import Union, List, ClassVar
from kqcircuits.simulations.export.solution import Solution


@dataclass
class ElmerSolution(Solution):
    """
    A Base class for Elmer Solution parameters

    Args:
        percent_error: Stopping criterion in adaptive meshing.
        max_error_scale: Maximum element error, relative to percent_error, allowed in individual elements.
        max_outlier_fraction: Maximum fraction of outliers from the total number of elements
        maximum_passes: Maximum number of adaptive meshing iterations.
        minimum_passes: Minimum number of adaptive meshing iterations.
        is_axisymmetric: Simulate with Axi Symmetric coordinates along :math:`y\\Big|_{x=0}` (Default: False)
        mesh_levels: If set larger than 1 Elmer will make the mesh finer by dividing each element
                     into 2^(dim) elements mesh_levels times. Default 1.
        mesh_size: Dictionary to determine mesh size where key (string) denotes material and value (double) denotes the
            maximal length of mesh element. Additional mesh size terms can be determined, if the value type is
            list. Then, term[0] is the maximal mesh element length inside at the entity and its expansion,
            term[1] is expansion distance in which the maximal mesh element length is constant (default=term[0]),
            and term[2] is the slope of the increase in the maximal mesh element length outside the entity.
            To refine material interface the material names by should be separated by '&' in the key. Key 'global_max'
            is reserved for setting global maximal element length. For example, if the dictionary is given as
            {'substrate': 10, 'substrate&vacuum': [2, 5], 'global_max': 100}, then the maximal mesh element length is 10
            inside the substrate and 2 on region which is less than 5 units away from the substrate-vacuum interface.
            Outside these regions, the mesh element size can increase up to 100.
        vtu_output: Output vtu files to view fields in Paraview.
                    Turning this off will make the simulations slightly faster

    """

    tool: ClassVar[str] = ""

    percent_error: float = 0.005
    max_error_scale: float = 2.0
    max_outlier_fraction: float = 1e-3
    maximum_passes: int = 1
    minimum_passes: int = 1
    is_axisymmetric: bool = False
    mesh_levels: int = 1
    mesh_size: dict = field(default_factory=dict)
    vtu_output: bool = True

    def get_solution_data(self):
        """Return the solution data in dictionary form."""
        sol_dict = {**self.__dict__, "tool": self.tool}
        sol_dict["solution_name"] = sol_dict.pop("name")
        return sol_dict


@dataclass
class ElmerVectorHelmholtzSolution(ElmerSolution):
    """
    Class for Elmer wave-equation solution parameters

    Args:
        frequency: Units are in GHz. Give a list of frequencies if using interpolating sweep.
        frequency_batch: Number of frequencies calculated between each round of fitting in interpolating sweep
        sweep_type: Type of frequency sweep. Options "explicit" and "interpolating".
        max_delta_s: Convergence tolerance in interpolating sweep
        london_penetration_depth: Allows supercurrent to flow on the metal boundaries within a layer
                                  of thickness `london_penetration_depth`
        quadratic_approximation: Use edge finite elements of second degree
        second_kind_basis: Use Nedelec finite elements of second kind

        use_av: Use a formulation of VectorHelmHoltz equation based on potentials A-V instead of electric field E.
                For details see https://www.nic.funet.fi/pub/sci/physics/elmer/doc/ElmerModelsManual.pdf
                WARNING: This option is experimental and might lead to poor convergence.
        conductivity: Adds a specified film conductivity on metal boundaries. Applies only when `use_av=True`
        nested_iteration: Enables alternative nested iterative solver to be used. Applies only when `use_av=True`
        convergence_tolerance: Convergence tolerance of the iterative solver. Applies only when `use_av=True`
        max_iterations: Maximum number of iterations for the iterative solver.
                        Applies only when `use_av=True` and only to the main solver (not to calc fields or port solver)
    """

    tool: ClassVar[str] = "wave_equation"

    frequency: Union[float, List[float]] = 5
    frequency_batch: int = 3
    sweep_type: str = "explicit"
    max_delta_s: float = 0.01
    london_penetration_depth: float = 0
    quadratic_approximation: bool = False
    second_kind_basis: bool = False

    use_av: bool = False
    conductivity: float = 0
    nested_iteration: bool = False
    convergence_tolerance: float = 1.0e-10
    max_iterations: int = 2000

    def __post_init__(self):
        """Cast frequency to list. Automatically called after init"""
        if isinstance(self.frequency, (float, int)):
            self.frequency = [self.frequency]
        elif not isinstance(self.frequency, list):
            self.frequency = list(self.frequency)


@dataclass
class ElmerCapacitanceSolution(ElmerSolution):
    """
    Class for Elmer capacitance solution parameters

    Args:
        p_element_order: polynomial order of p-elements
        linear_system_method: Options: 1. mg (multigrid), 2. bicgstab or any other iterative solver mentioned in
                              ElmerSolver manual section 4.3.1
        integrate_energies: Calculate energy integrals over each object. Used in EPR simulations
        convergence_tolerance: Convergence tolerance of the iterative solver.
        max_iterations: Maximum number of iterations for the iterative solver.
    """

    tool: ClassVar[str] = "capacitance"

    p_element_order: int = 3
    linear_system_method: str = "mg"
    integrate_energies: bool = False
    convergence_tolerance: float = 1.0e-9
    max_iterations: int = 500


@dataclass
class ElmerCrossSectionSolution(ElmerSolution):
    """
    Class for Elmer cross-section solution parameters.
    By default both 2D Capacitance and 2D Inductance simulation will be run when using this

    Args:
        p_element_order: polynomial order of p-elements
        linear_system_method: Options: 1. mg (multigrid), 2. bicgstab or any other iterative solver mentioned in
                              ElmerSolver manual section 4.3.1
        integrate_energies: Calculate energy integrals over each object. Used in EPR simulations
        boundary_conditions: Parameters to determine boundary conditions for potential on the edges
                             of simulation box. Supported keys are `xmin` , `xmax` ,`ymin` and `ymax`
                             Example: `boundary_conditions = {"xmin": {"potential": 0}}`
        convergence_tolerance: Convergence tolerance of the iterative solver.
                               Applies only to capacitance part of the simulation
        max_iterations: Maximum number of iterations for the iterative solver.
                        Applies only to capacitance part of the simulation
        run_inductance_sim: Can be used to skip running the inductance simulation and just do 2D capacitance.
                            No impendance can then be calculated but useful for making EPR simulations faster
    """

    tool: ClassVar[str] = "cross-section"

    p_element_order: int = 3
    linear_system_method: str = "mg"
    integrate_energies: bool = False
    boundary_conditions: dict = field(default_factory=dict)
    convergence_tolerance: float = 1.0e-9
    max_iterations: int = 500
    run_inductance_sim: bool = True


@dataclass
class ElmerEPR3DSolution(ElmerSolution):
    """
    Class for Elmer 3D EPR simulations. Similar to electrostatics simulations done with ElmerCapacitanceSolution,
    but supports separating energies by PartitionRegions and produces no capacitance matrix

    Args:
        p_element_order: polynomial order of p-elements
        linear_system_method: Options: 1. mg (multigrid), 2. bicgstab or any other iterative solver mentioned in
                              ElmerSolver manual section 4.3.1
        convergence_tolerance: Convergence tolerance of the iterative solver.
        max_iterations: Maximum number of iterations for the iterative solver.
    """

    tool: ClassVar[str] = "epr_3d"

    p_element_order: int = 3
    linear_system_method: str = "bicgstab"
    convergence_tolerance: float = 1.0e-9
    max_iterations: int = 1000


def get_elmer_solution(tool="capacitance", **solution_params):
    """Returns an instance of ElmerSolution subclass.

    Args:
        tool: Determines the subclass of ElmerSolution.
        solution_params: Arguments passed for  ElmerSolution subclass.
    """
    for c in [ElmerVectorHelmholtzSolution, ElmerCapacitanceSolution, ElmerCrossSectionSolution, ElmerEPR3DSolution]:
        if tool == c.tool:
            return c(**solution_params)
    raise ValueError(f"No ElmerSolution found for tool={tool}.")



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

"""
Functions to tune and replace junctions in existing design files.

See scripts/macros/export/export_tuned_junctions.lym for a use case of these functions
"""

from os import path
from typing import Dict, List
import logging
from kqcircuits.pya_resolver import pya
from kqcircuits.junctions import junction_type_choices
from kqcircuits.junctions.junction import Junction
from kqcircuits.chips.chip import Chip
from kqcircuits.util.library_helper import load_libraries, to_library_name


class JunctionEntry:
    """All junction properties we want to store when extracting junctions"""

    def __init__(self, class_type: type, trans: pya.DCplxTrans, parameters: Dict, parent_name: str, name: str) -> None:
        self.type = class_type
        self.trans = trans
        self.parameters = parameters
        self.parent_name = parent_name
        self.name = name

    def __eq__(self, __value: object) -> bool:
        return (
            self.type == __value.type
            and self.trans == __value.trans
            and self.parameters == __value.parameters
            and self.parent_name == __value.parent_name
            and self.name == __value.name
        )


def _check_junction_names_unique(junctions):
    """Raises exception if ``junctions`` contains a non-unique (parent_name, name) key-pair"""
    unique_names = set()
    for junction in junctions:
        if (junction.parent_name, junction.name) in unique_names:
            error_text = (
                "Following cell parent name and child name is not unique in top cell: "
                f"{(junction.parent_name, junction.name)}. "
                "Something seems to be wrong with KQC generated cell"
            )
            error = ValueError(error_text)
            logging.exception(error_text, exc_info=error)
            raise error
        unique_names.add((junction.parent_name, junction.name))


def _check_missing_junction_parameters(
    junction_class_name, junction_schema_errors, params, tuned_params, parent_name, name
):
    """Run for every found junction to compare junction library schema (`params`)
    and schema as given in json file (`tuned_params`).

    If some parameter keys are missing in `tuned_params`, stores missing keys in mutable `junction_schema_errors`.
    `junction_schema_errors` is a dict that has junction class as keys, and as values
    a tuple "missing_fields" with set of missing parameter keys and a list for junction names.

    Note that some parameter keys are ignored.
    """
    ignore_param_keys = {"_junction_parameters", "junction_parameters", "display_name"}
    if junction_class_name not in junction_schema_errors:
        junction_schema_errors[junction_class_name] = {"missing_fields": (set(), []), "surplus_fields": (set(), [])}
    missing_fields = set(params.keys()).difference(set(tuned_params.keys())).difference(ignore_param_keys)
    junction_schema_errors[junction_class_name]["missing_fields"][0].update(missing_fields)
    if len(missing_fields) > 0:
        junction_schema_errors[junction_class_name]["missing_fields"][1].append((parent_name, name))


def _check_surplus_junction_parameters(
    junction_class_name, junction_schema_errors, params, tuned_params, parent_name, name
):
    """Run for every found junction to compare junction library schema (`params`)
    and schema as given in json file (`tuned_params`).

    If some parameter keys are in `tuned_params` but not in `params`,
    stores missing keys in mutable `junction_schema_errors`.
    `junction_schema_errors` is a dict that has junction class as keys, and as values
    a tuple "surplus_fields" with set of surplus parameter keys and a list for junction names.
    """
    if junction_class_name not in junction_schema_errors:
        junction_schema_errors[junction_class_name] = {"missing_fields": (set(), []), "surplus_fields": (set(), [])}
    surplus_fields = set(tuned_params.keys()).difference(set(params.keys()))
    junction_schema_errors[junction_class_name]["surplus_fields"][0].update(surplus_fields)
    if len(surplus_fields) > 0:
        junction_schema_errors[junction_class_name]["surplus_fields"][1].append((parent_name, name))


def _print_surplus_junction_parameters(junction_schema_errors):
    """Logs as warning the content of "surplus_fields" in `junction_schema_errors`
    junction parameters that were attempted to be tuned,
    yet were not defined for the given junction types.
    """
    for k, v in junction_schema_errors.items():
        surplus_fields, junctions = v["surplus_fields"]
        if len(surplus_fields) > 0:
            logging.warning(
                (
                    f"{k} class junction attempted to be tuned with parameters "
                    f"that are not part of the class: {surplus_fields}"
                )
            )
            logging.warning(f"for {junctions[:5]}\n")


def _halt_if_missing_junction_parameters(junction_schema_errors, is_pcell):
    """Raises exception if "missing_fields" for some entry in `junction_schema_errors` is not empty,
    formats an error message to show all missing parameter keys detected for each junction type
    and names the affected junctions.
    """
    error_text = ""
    for k, v in junction_schema_errors.items():
        missing_fields, junctions = v["missing_fields"]
        if len(missing_fields) > 0:
            error_text = (
                f"{error_text}"
                f"{k} class junction parameters missing {missing_fields}\n"
                f"missing for {junctions[:5]}\n\n"
            )
    if len(error_text) > 0:
        if is_pcell:
            error_text = (
                "Since junction type was changed for some junctions, "
                "the tuned junction json should give value at least for parameters "
                "that are in new junction type but not in old junction type.\n"
                f"Following junction parameters missing:\n\n{error_text}"
            )
        else:
            error_text = (
                "Since the cell doesn't contain pre-existing PCell parameter data, "
                "the tuned junction json should be exhaustive.\n"
                f"Following junction parameters missing:\n\n{error_text}"
            )
        error = ValueError("Some junction parameters were missing in the tuning json, see log for details")
        logging.exception(error_text, exc_info=error)
        raise error


def extract_junctions(top_cell: pya.Cell, tuned_junction_parameters: Dict) -> List[JunctionEntry]:
    """Extracts all junction elements placed in the `top_cell`.
    Junction parameters are tuned according to `tuned_junction_parameters` dict.

    `tuned_junction_parameters` is a dict with junction's parent cell name as key,
    where parent cell is an element that contains the junction, e.g. "QB1", "testarray_nw" etc.
    The value is also a dict, with junction cell's name as key, e.g "squid", "squid_0", "squid_3" etc.
    For example testarray cells may have multiple junction cells.
    Then `tuned_junction_parameters[parent_name][name]` is a dict of junction parameters.

    If `top_cell` has pcell data, the parameter values that are missing in `tuned_junction_parameters`
    can be inferred from the Junction PCell's values. So `tuned_junction_parameters` may only contain
    parameter values that are different from how junctions were defined in `top_cell`.

    If `top_cell` has no pcell data, `tuned_junction_parameters` must include all parameter keys
    of the junction parameter schema for each junction contained in the `top_cell`,
    even if the parameter values are the same as were used to construct `top_cell`.
    If that is not the case, `extract_junctions` will raise an exception.

    Returns a list of `JunctionEntry` objects that can be used to place the extracted junctions
    into another cell that has tuned parameters but is otherwise identical in shape, placement and orientation.

    Junction type may also be changed, if `junction_type` is tuned to have some other junction class name.
    For every junction that has its `junction_type` changed, even if the cell contains PCell data,
    `tuned_junction_parameters` should have at least all parameters present that are in the new junction type
    but not in the old junction type.
    """
    junction_schema_errors = {}
    found_junctions = []
    library_layout = (load_libraries(path=Junction.LIBRARY_PATH)[Junction.LIBRARY_NAME]).layout()
    layout = top_cell.layout()
    is_pcell = False
    for i in top_cell.each_inst():
        if i.pcell_declaration() is not None:
            is_pcell = True
            break
    if not is_pcell:
        logging.warning("Top cell doesn't contain PCell parameter data")

    def recursive_junction_search(inst, parent_name, prev_trans):
        cell = layout.cell(inst.cell_index)
        name = inst.property("id")
        trans = prev_trans * inst.dcplx_trans
        tuned_params = tuned_junction_parameters.get(parent_name, {}).get(name, {})
        if is_pcell:
            pcell = inst.pcell_declaration()
            is_junction = pcell and isinstance(pcell, Junction)
        else:
            cell_class_from_name = cell.name.split("$")[0].replace("*", " ")
            pcell = library_layout.pcell_declaration(cell_class_from_name)
            is_junction = cell_class_from_name in junction_type_choices
        if is_junction:
            if is_pcell:
                # Parameter values present in PCell data can be reused
                pcell_param_values = inst.pcell_parameters_by_name()
            junction_type = tuned_params.get("junction_type")
            if junction_type not in junction_type_choices and not (is_pcell and "junction_type" not in tuned_params):
                error_text = (
                    f"'junction_type' value {junction_type} for junction "
                    f"({parent_name}, {name}) is not part of junction_type_choices"
                )
                error = ValueError(error_text)
                logging.exception(error_text, exc_info=error)
                raise error
            if not junction_type and is_pcell:
                junction_type = pcell_param_values.get("junction_type")
            junction_type = library_layout.pcell_declaration(junction_type)
            params = {
                # If PCell is available, get PCell parameter values that are available
                k: v.default if not is_pcell else pcell_param_values.get(k, v.default)
                for k, v in type(junction_type).get_schema().items()
            }
            params.update(tuned_params)
            # Not PCell, need to be strict that tuned junction params json includes all params
            if not is_pcell:
                _check_missing_junction_parameters(
                    type(junction_type).__name__, junction_schema_errors, params, tuned_params, parent_name, name
                )
            # Is PCell, and junction type is being changed. Need to make sure params exclusive to new type are tuned
            elif junction_type is not None and junction_type != pcell:
                exclusive_params = {k: v for k, v in params.items() if k not in pcell_param_values}
                _check_missing_junction_parameters(
                    type(junction_type).__name__,
                    junction_schema_errors,
                    exclusive_params,
                    tuned_params,
                    parent_name,
                    name,
                )
            _check_surplus_junction_parameters(
                type(junction_type).__name__,
                junction_schema_errors,
                type(junction_type).get_schema(),
                tuned_params,
                parent_name,
                name,
            )
            found_junctions.append(JunctionEntry(type(junction_type), trans, params, parent_name, name))
        for i in cell.each_inst():
            # For pcell oas, accumulate transformation starting from root
            # For static oas, only use parent.dcplx_trans * this.dcplx_trans
            recursive_junction_search(i, name, trans if is_pcell else prev_trans)

    for i in top_cell.each_inst():
        recursive_junction_search(i, None, i.dcplx_trans)
    _check_junction_names_unique(found_junctions)
    _print_surplus_junction_parameters(junction_schema_errors)
    _halt_if_missing_junction_parameters(junction_schema_errors, is_pcell)
    return found_junctions


def place_junctions(top_cell: pya.Cell, junctions: List[JunctionEntry]) -> None:
    """Places `junctions` to `top_cell` in the same location and orientation as in
    the cell they were extracted from, but with possibly tuned parameters.
    """
    layout = top_cell.layout()
    for junction in junctions:
        if "junction_type" not in junction.parameters:
            junction.parameters["junction_type"] = to_library_name(junction.type.__name__)
        if to_library_name(junction.type.__name__) != junction.parameters["junction_type"]:
            error_text = (
                f"Exported junction of class '{to_library_name(junction.type.__name__)}', "
                f"but 'junction_type' parameter was set to {junction.parameters['junction_type']}"
            )
            error = ValueError(error_text)
            logging.exception(error_text, exc_info=error)
            raise error
        junction_cell = Junction.create(layout, **junction.parameters)
        top_cell.insert(pya.DCellInstArray(junction_cell.cell_index(), junction.trans))


def get_tuned_junction_json(junctions: List[JunctionEntry]) -> Dict:
    """Returns a jsonable dict of all junction parameters for each junction entry in `junctions`.

    If junctions were extracted from a cell with pcell data, the json can be extracted to
    have an exhaustive list of all junction parameters so that junctions can then be tuned
    using a cell with no pcell data, which is faster to read.
    """
    result = {}
    for junction in junctions:
        if junction.parent_name not in result:
            result[junction.parent_name] = {}
        result[junction.parent_name][junction.name] = junction.parameters
    return result


def copy_one_layer_of_cell(
    write_path: str, top_cell: pya.Cell, junctions: List[JunctionEntry], layer_string: str
) -> None:
    """Extracts all geometry in `top_cell` at layer `layer_string`
    and saves the geometry into a new file at `write_path`.
    The face of the layer is determined from `junctions` parameters.

    This can be used to extract geometry of alignment markers as well as other geometry
    to visualize junctions within a context of surrounding elements.
    The file at `write_path` may be loaded later and junctions may be placed using
    `place_junctions` into the top cell of the file, then saved again.
    """
    # TODO: Assuming face of the junction determined by first element of 'face_ids'.
    # Reconsider once multiface junctions are introduced.
    faces_set = {j.parameters["face_ids"][0] for j in junctions}
    if len(faces_set) > 1:
        error_text = f"Detected inconsistent junction face assignments {faces_set}"
        error = ValueError(error_text)
        logging.exception(error_text, exc_info=error)
        raise error
    face = list(faces_set)[0]
    layout = top_cell.layout()
    layers = [l for l in layout.layer_infos() if l.name == (f"{face}_{layer_string}")]
    if not layers:
        error_text = f"Layer not found '{face}_{layer_string}'"
        error = ValueError(error_text)
        logging.exception(error_text, exc_info=error)
        raise error
    layer = layers[0]

    svopt = pya.SaveLayoutOptions()
    svopt.set_format_from_filename(write_path)
    svopt.deselect_all_layers()
    svopt.clear_cells()
    svopt.add_cell(top_cell.cell_index())
    svopt.add_layer(layout.layer(layer), layer)
    svopt.write_context_info = False
    layout.write(write_path, svopt)


def replace_squids(cell, junction_type, parameter_name, parameter_start, parameter_step, parameter_end=None):
    """DEPRECATED! Replaces squids by code generated squids with the given parameter sweep.

    All squids below top_cell in the cell hierarchy are removed. The number of code
    generated squids may be limited by the value of parameter_end.

    Args:
        cell (Cell): The cell where the squids to be replaced are
        junction_type: class name of the code generated squid that replaces the other squids
        parameter_name (str): Name of the parameter to be swept
        parameter_start: Start value of the parameter
        parameter_step: Parameter value increment step
        parameter_end: End value of the parameter. If None, there is no limit for the parameter value, so that all
            squids are replaced

    """
    layout = cell.layout()
    parameter_value = parameter_start
    junction_types = [choice if isinstance(choice, str) else choice[1] for choice in junction_type_choices]

    old_squids = []  # list of tuples (squid instance, squid dtrans with respect to cell, old name)

    def recursive_replace_squids(top_cell_inst, combined_dtrans):
        """Appends to old_squids all squids in top_cell_inst or any instance below it in hierarchy."""
        # cannot use just top_cell_inst.cell due to klayout bug, see
        # https://www.klayout.de/forum/discussion/1191/cell-shapes-cannot-call-non-const-method-on-a-const-reference
        top_cell = layout.cell(top_cell_inst.cell_index)
        for subcell_inst in top_cell.each_inst():
            subcell_name = subcell_inst.cell.name.split("$")[0]
            if subcell_name in junction_types:
                old_squids.append((subcell_inst, combined_dtrans * subcell_inst.dcplx_trans, subcell_name))
            else:
                recursive_replace_squids(subcell_inst, combined_dtrans * subcell_inst.dcplx_trans)

    for inst in cell.each_inst():
        if inst.cell.name in junction_types:
            old_squids.append((inst, inst.dcplx_trans, inst.cell.name))
        recursive_replace_squids(inst, inst.dcplx_trans)

    # sort left-to-right and bottom-to-top
    old_squids.sort(key=lambda squid: (squid[1].disp.x, squid[1].disp.y))

    for inst, dtrans, name in old_squids:
        if (parameter_end is None) or (parameter_value <= parameter_end):
            # create new squid at old squid's position
            parameters = {parameter_name: parameter_value}
            squid_cell = Junction.create(
                layout, junction_type=junction_type, face_ids=inst.pcell_parameter("face_ids"), **parameters
            )
            cell.insert(pya.DCellInstArray(squid_cell.cell_index(), dtrans))
            logging.info(
                'Replaced squid "%s" with dtrans=%s by a squid "%s" with %s=%s.',
                name,
                dtrans,
                junction_type,
                parameter_name,
                parameter_value,
            )
            parameter_value += parameter_step
        # delete old squid
        inst.delete()


def replace_squid(top_cell, inst_name, junction_type, mirror=False, squid_index=0, **params):
    """DEPRECATED! Replaces a SQUID by the requested alternative in the named instance.

    Replaces the SQUID(s) in the sub-element(s) named ``inst_name`` with other SQUID(s) of
    ``junction_type``. The necessary SQUID parameters are specified in ``params``. If ``inst_name`` is
    a Test Structure then ``squid_index`` specifies which SQUID to change.

    Args:
        top_cell: The top cell with SQUIDs to be replaced
        inst_name: Instance name of PCell containing the SQUID to be replaced
        junction_type: Name of SQUID Class or .gds/.oas file
        mirror: Mirror the SQUID along its vertical axis
        squid_index: Index of the SQUID to be replaced within a Test Structure
        **params: Extra parameters for the new SQUID
    """

    def find_cells_with_squids(chip, inst_name):
        """Returns the container cells in `chip` called `inst_name`"""
        cells = []
        layout = chip.layout()
        for inst in chip.each_inst():
            if inst.property("id") == inst_name:
                cells.append((chip, inst))
            elif isinstance(inst.pcell_declaration(), Chip):  # recursively look for more chips
                cells += find_cells_with_squids(layout.cell(inst.cell_index), inst_name)
        return cells

    cells = find_cells_with_squids(top_cell, inst_name)
    if not cells:
        logging.warning(f"Could not find anything named '{inst_name}'!")

    layout = top_cell.layout()
    file_cell = None
    if junction_type.endswith(".oas") or junction_type.endswith(".gds"):  # try to load from file
        if not path.exists(junction_type):
            logging.warning(f"No file found at '{path.realpath(junction_type)}!")
            return
        load_opts = pya.LoadLayoutOptions()
        load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
        layout.read(junction_type, load_opts)
        file_cell = layout.top_cells()[-1]
        file_cell.name = f"Junction Library.{file_cell.name}"

    for chip, inst in cells:
        orig_trans = inst.dcplx_trans
        ccell = inst.layout().cell(inst.cell_index)

        if ccell.is_pcell_variant():  # make copy if used elsewhere
            dup = ccell.dup()
            dup.set_property("id", inst.property("id"))
            inst.delete()
            chip.insert(pya.DCellInstArray(dup.cell_index(), orig_trans), dup.prop_id)
            ccell = dup

        squids = [sq for sq in ccell.each_inst() if sq.cell.qname().find("Junction Library") != -1]
        squids.sort(key=lambda q: q.property("squid_index"))
        if not squids or squid_index >= len(squids) or squid_index < 0:
            logging.warning(f"No SQUID found in '{inst_name}' or squid_index={squid_index} is out of range!")
            continue
        old_squid = squids[squid_index]
        if old_squid.is_pcell():
            params = {"face_ids": old_squid.pcell_parameter("face_ids"), **params}
        trans = old_squid.dcplx_trans * pya.DCplxTrans.M90 if mirror else old_squid.dcplx_trans
        squid_pos = (orig_trans * trans).disp
        logging.info(f"Replaced SQUID of '{inst_name}' with {junction_type} at {squid_pos}.")
        old_squid.delete()
        if file_cell:
            new_squid = ccell.insert(pya.DCellInstArray(file_cell.cell_index(), trans))
        else:
            new_squid = Junction.create(layout, junction_type=junction_type, **params)
            new_squid = ccell.insert(pya.DCellInstArray(new_squid.cell_index(), trans))
        new_squid.set_property("squid_index", squid_index)


def convert_cells_to_static(layout):
    """DEPRECATED! Converts all cells in the layout to static."""

    converted_cells = {}

    # convert the cells to static
    for cell in layout.each_cell():
        if cell.is_library_cell():
            cell_idx = cell.cell_index()
            new_cell_idx = layout.convert_cell_to_static(cell_idx)
            if new_cell_idx != cell_idx:
                converted_cells[cell_idx] = new_cell_idx

    # translate the instances
    for cell in layout.each_cell():
        for inst in cell.each_inst():
            if inst.cell_index in converted_cells:
                inst.cell_index = converted_cells[inst.cell_index]

    # delete the PCells
    for cell_idx in converted_cells:
        layout.delete_cell(cell_idx)



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import os
from pathlib import Path


def find_ansys_executable(default):
    """Finds latest Ansys Electronics executable from default installation locations. Returns 'default' if not found."""
    paths = [
        (Path(os.environ.get("ProgramFiles", r"C:\Program Files")).joinpath("AnsysEM"), r"Win64\ansysedt.exe"),
        (Path("/opt/AnsysEM"), "Linux64/ansysedt"),
    ]
    for root, exe in paths:
        if os.path.isdir(root):
            versions = sorted([f for f in os.listdir(root) if f.startswith("v")], reverse=True)
            for version in versions:
                executable = root.joinpath(version).joinpath(exe)
                if os.path.isfile(executable):
                    return executable
    return default



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import subprocess
from kqcircuits.defaults import ROOT_PATH


def export_singularity(remote_host: str, singularity_remote_path: str):

    if singularity_remote_path is None:
        singularity_remote_path = "~/KQCircuits/singularity"

    subprocess.call(
        [
            "ssh",
            remote_host,
            "mkdir",
            "-p",
            singularity_remote_path,
            singularity_remote_path + "/libexec",
            singularity_remote_path + "/bin",
        ]
    )

    subprocess.call(
        [
            "scp",
            ROOT_PATH / "singularity/libexec/kqclib",
            ROOT_PATH / "singularity/libexec/kqclib.sh",
            remote_host + ":" + singularity_remote_path + "/libexec",
        ]
    )

    subprocess.call(["scp", ROOT_PATH / "singularity/create_links.sh", remote_host + ":" + singularity_remote_path])

    subprocess.call(
        [
            "ssh",
            remote_host,
            "cd",
            singularity_remote_path,
            "&&",
            "./create_links.sh",
            "&&",
            "mv",
            "python",
            "bin/python",
            "&&",
            "rm",
            "bin/paraview",
        ]
    )



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
from pathlib import Path
import subprocess
import platform
import sys
import logging

from kqcircuits.defaults import EXPORT_PATH_IDENTIFIER

logging.basicConfig(level=logging.WARN, stream=sys.stdout)


def export_and_run(export_script: Path, export_path: Path, quiet: bool = False, export_only: bool = False, args=None):
    """
    Exports and runs a KQC simulation.

    Args:
        export_script(Path): path to the simulation export script
        export_path(Path): path where simulation files are exported set with `--export-path-basename`. If None, the
                           path set in export script will be used.
        quiet(bool): if True all the GUI dialogs are shown, otherwise not.
        export_only(bool): if True no simulation is run, only export files.
        args(list): a list of strings describing arguments to be passed to the simulation script

    Returns:
        a tuple containing

            * export_script(Path): path to the simulation export script
            * export_path(list(Path)): list of paths where simulation files are exported

    """

    script_export_paths = run_export_script(export_script, export_path, quiet, args)

    if not export_only:
        run_simulations(script_export_paths)

    return export_script, script_export_paths


def run_export_script(export_script: Path, export_path: Path, quiet: bool = False, args=None):
    """
    Generate the simulation files by running the export script. Returns list of paths where
    simulation files are exported. Returned paths are parsed from stdout of the export script printed
    by function `create_or_empty_tmp_directory`, based on the identifier `EXPORT_PATH_IDENTIFIER`
    """
    if args is None:
        args = []
    elif "--simulation-export-path" in args:
        logging.error("--simulation-export-path is not allowed!")
        sys.exit()

    export_cmd = (
        [sys.executable, export_script]
        + (["--simulation-export-path", str(export_path)] if export_path else [])
        + args
        + (["-q"] if quiet else [])
    )
    # Run export script and capture stdout to be processed
    with subprocess.Popen(export_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True) as process:
        process_stdout, process_stderr = process.communicate()
        print(process_stdout)
        print(process_stderr, file=sys.stderr)
        if process.returncode:
            raise subprocess.CalledProcessError(process.returncode, export_cmd)

    # Parse export paths from stdout printed in `create_or_empty_tmp_directory`
    script_export_paths = [l.strip() for l in process_stdout.split("\n")]
    script_export_paths = [
        Path(l.removeprefix(EXPORT_PATH_IDENTIFIER))
        for l in script_export_paths
        if l.startswith(EXPORT_PATH_IDENTIFIER)
    ]

    # remove duplicate paths
    unique_paths = set()
    script_export_paths = [n for n in script_export_paths if not (n in unique_paths or unique_paths.add(n))]

    if export_path and len(script_export_paths) > 1:
        logging.error("Using `--export-path-basename` is not supported with scripts exporting multiple simulations")

    return script_export_paths


def run_simulations(script_export_paths: list[Path]):
    """Run exported simulations"""
    for script_export_path in script_export_paths:
        if (script_export_path / "simulation.sh").is_file():
            simulation_shell_script = "simulation.sh"
        elif (script_export_path / "simulation.bat").is_file():
            simulation_shell_script = "simulation.bat"
        else:
            logging.warning(f"No simulation.sh or .bat script found in {script_export_path}")
            continue

        if platform.system() == "Windows":  # Windows
            subprocess.call(simulation_shell_script, shell=True, cwd=str(script_export_path))
        elif platform.system() == "Darwin":  # macOS
            subprocess.call(["bash", simulation_shell_script], cwd=str(script_export_path))
        else:  # Linux
            subprocess.call(["bash", simulation_shell_script], cwd=str(script_export_path))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from kqcircuits.chips.chip import Chip
from kqcircuits.defaults import default_sampleholders
from kqcircuits.util.parameters import Param, pdt

sampleholder_type_choices = list(default_sampleholders.keys())


class Launchers(Chip):
    """The PCell declaration for a chip with predefined launchers."""

    sampleholder_type = Param(pdt.TypeString, "Type of the launchers", "SMA8", choices=sampleholder_type_choices)

    def build(self):
        self.produce_launchers(self.sampleholder_type)



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.simulations.export.ansys.ansys_solution import (
    AnsysEigenmodeSolution,
    AnsysCurrentSolution,
    AnsysVoltageSolution,
    AnsysHfssSolution,
)
from kqcircuits.simulations.export.elmer.elmer_solution import (
    ElmerVectorHelmholtzSolution,
    ElmerCapacitanceSolution,
    ElmerCrossSectionSolution,
)


def validate_simulation(simulation, solution):
    """Analyses Simulation and Solution objects and raises an error if inconsistencies in configuration are found.

    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    simulation_and_solution_types_match(simulation, solution)
    has_no_ports_when_required(simulation, solution)
    has_edgeport_when_forbidden(simulation, solution)
    flux_integration_layer_exists_if_needed(simulation, solution)


def simulation_and_solution_types_match(simulation, solution):
    """Validation check: ensures that a simulation and solution types match.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    if isinstance(simulation, CrossSectionSimulation) != isinstance(solution, ElmerCrossSectionSolution):
        raise ValidateSimError(
            f"Simulation '{simulation.name}' is incompatible with {type(solution)}",
            validation_type=simulation_and_solution_types_match.__name__,
        )


def has_no_ports_when_required(simulation, solution):
    """Validation check: ensures that a simulation object has ports when the solution type requires it.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    port_names = get_port_names(simulation)
    sim_name = simulation.name
    if not port_names and isinstance(
        solution,
        (
            AnsysHfssSolution,
            AnsysVoltageSolution,
            AnsysCurrentSolution,
            ElmerVectorHelmholtzSolution,
            ElmerCapacitanceSolution,
        ),
    ):
        raise ValidateSimError(
            f"Simulation '{sim_name}' has no ports assigned. This is incompatible with {type(solution)}",
            validation_type=has_no_ports_when_required.__name__,
        )


def has_edgeport_when_forbidden(simulation, solution):
    """Validation check: ensure that if at least one "EdgePort" is present, some solution types can't be chosen.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    port_names = get_port_names(simulation)
    sim_name = simulation.name
    if "EdgePort" in port_names and isinstance(
        solution,
        (
            AnsysEigenmodeSolution,
            AnsysVoltageSolution,
            AnsysCurrentSolution,
        ),
    ):
        raise ValidateSimError(
            f"Simulation '{sim_name}' has at least one 'EdgePort'. This is incompatible with {type(solution)}",
            validation_type=has_edgeport_when_forbidden.__name__,
        )


def flux_integration_layer_exists_if_needed(simulation, solution):
    """Validation check related to the presence of layers and magnetic flux integration.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    sim_name = simulation.name
    has_integrate_flux = hasattr(solution, "integrate_magnetic_flux")
    integrate_flux = solution.integrate_magnetic_flux if has_integrate_flux else False

    # Ensures that a layer with thickness == 0 and a material != "pec"
    # exists in the setup when "integrate_magnetic_flux" is True.
    if integrate_flux:
        layers = simulation.layers
        has_flux_integration_layer = False
        for layer in layers.values():
            if layer.get("thickness", -1) == 0 and layer.get("material", "") != "pec":
                has_flux_integration_layer = True
                break
        if not has_flux_integration_layer:
            raise ValidateSimError(
                f"Simulation '{sim_name}' has 'integrate_magnetic_flux = True' "
                + "but the integration layer is missing.",
                validation_type=flux_integration_layer_exists_if_needed.__name__,
            )


# Following code is not validation checks but utilities used by validity checks


def get_port_names(simulation):
    """Helper function that returns a list of port names in a Simulation object.
    Args:
        simulation: A Simulation object.
    Returns:
        port_names: A list of names related to the ports present in simulation.
    """
    port_list = simulation.ports if isinstance(simulation, Simulation) else []
    port_names = []
    for port in port_list:
        port_names.append(type(port).__name__)
    return port_names


class ValidateSimError(Exception):
    """Custom exception class for specific error handling."""

    def __init__(self, message, validation_type=None):
        super().__init__(message)
        self.validation_type = validation_type
